# -*- coding: utf-8 -*-
"""Simple code to fit critical temperatures in R(T) data."""
# pylint: disable=invalid-name
import numpy as np

from Stoner import Data
from Stoner.analysis.fitting.models.generic import linear


filename = (
    "../sample-data/6221-Lockin-DAQ Temperature Control -30.0Deg 0.004T.txt"
)

t_col = ": T2"  # Temperature column label
r_cols = ("Nb::R", "YIG-Nb-Co::R")  # Resistance Column Labels
iterator = "iterator"  # Temperature ramp iteration column label
threshold = 0.85  # Fraction of transition to fit to

data = Data(
    filename
)  # Use FALSE to get a dialog box to the file containing Tc data

# Define my working x and y axes
# Split one file into a folder of two files by the iterator column
fldr = data.split(iterator)

result = Data()
for data in fldr:  # For each iteration ramp in the Tc data
    if len(data) == 0:
        continue
    row = [data.mean(iterator)]
    data.figure(figsize=(8, 4))
    for i, r_col in enumerate(r_cols):
        data.setas(x=t_col, y=r_col)
        data.del_rows(~np.isnan(data.y))

        # Normalise data on y axis between +/- 1
        data.normalise(base=(-1.0, 1.0), replace=True)

        # Swap x and y axes around so that R is x and T is y
        data = ~data

        # Curve fit a straight line, using only the central 90% of the resistance transition
        data.curve_fit(
            linear,
            bounds=lambda x, r: -threshold < x < threshold,
            result=True,
            p0=[7.0, 0.0],
        )  # result=True to record fit into metadata

        # Plot the results
        data.setas[-1] = "y"
        data.subplot(1, len(r_cols), i + 1)
        data.plot(fmt=["k.", "r-"])
        data.annotate_fit(linear, x=0.1, y=0.1, fontsize="small")
        data.title = f"Ramp {row[0]}"
        row.extend([data["linear:intercept"], data["linear:intercept err"]])
    data.tight_layout()
    result += np.array(row)

result.column_headers = ["Ramp", "Sample 4 R", "dR", "Sample 7 R", "dR"]
result.setas = "xyeye"
result.plot(fmt=["k.", "r."])



"""Customising a template for plotting."""

# pylint: disable=invalid-name
from Stoner import Data
from Stoner.plot.formats import SketchPlot

p = Data("sample.txt", setas="xy", template=SketchPlot)
p.plot()



"""Demo of new Stoner.Analysis.AnalyseFile.lmfit."""

# pylint: disable=invalid-name
import urllib
import io

from Stoner.analysis.fitting.models import cfg_data_from_ini, cfg_model_from_ini

config = io.StringIO(
    urllib.request.urlopen(  # pylint: disable=urllib_urlopen
        "https://raw.githubusercontent.com/stonerlab/Stoner-PythonCode/master/scripts/PCAR-chi%5E2.ini"
    )
    .read()
    .decode()
)
datafile = io.StringIO(  # pylint: disable=urllib_urlopen
    urllib.request.urlopen(
        "https://github.com/stonerlab/Stoner-PythonCode/raw/master/sample-data/PCAR%20Co%20Data.csv"
    )
    .read()
    .decode()
)

d = cfg_data_from_ini(config, datafile)
model, p0 = cfg_model_from_ini(config, data=d)

fit = d.lmfit(model, result=True, header="lmfit", output="report")

d.plot_xy(0, [2, 1], ["r-", "bo"], title="PCAR Co Data")  # plot the data
d.annotate_fit(model, x=0.05, y=0.75, fontdict={"size": "x-small"})

print(fit.fit_report())



# -*- coding: utf-8 -*-
"""Functions used by the AnalysisMixin class."""

import numpy as np
from scipy.interpolate import interp1d
from scipy.optimize import curve_fit, newton
from scipy.signal import get_window

__all__ = ["outlier", "threshold", "_twoD_fit", "ApplyAffineTransform", "GetAffineTransform", "poly_outlier"]


def outlier(row, window, metric, ycol=None, shape="bopxcar"):
    """Outlier detector function.

    Calculates if the current row is an outlier from the surrounding data by looking
    at the number of standard deviations away from the average of the window it is.

    Args:
        row (array):
            Single row of the dataset.
        window (array):
            Section of data surrounding the row being examined.
        metric (float):
            distance the current row is from the local mean.

    Keyword Arguments:
        ycol (column index or None):
            If set, specifies the column containing the data to check.

    Returns:
        (bool):
            If True, then the current row is an outlier from the local data.
    """
    windowing = get_window(shape, len(window))
    windowing /= windowing.sum() / windowing.size
    av = np.average(window[:, ycol] * windowing)
    std = np.std(window[:, ycol])  # standard deviation
    return abs(row[ycol] - av) > metric * std


def poly_outlier(row, window, metric=3.0, ycol=None, xcol=None, order=1, yerr=None):
    """Alternative outlier detection function that fits a polynomial locally over the window.

    Args:
        row (1D array):
            Current row of data
        column int):
            Column index of y values to examine
        window (2D array):
            Local window of data

    Keyyword Arguments:
        metric (float):
            Some measure of how sensitive the dection should be
        xcol (column index):
            Column of data to use for X values. Defaults to current setas value
        order (int):
            Order of polynomial to fit. Must be < length of window-1

    Returns:
        (bool):
            True if current row is an outlier
    """
    if order > window.shape[0] - 2:
        raise ValueError(f"order should be smaller than the window length. {order} vs {window.shape[0] - 2}")

    x = window[:, xcol] - row[xcol]
    y = window[:, ycol]
    if yerr:
        w = 1.0 / window[:, yerr]
    else:
        w = None

    popt, pcov = np.polyfit(x, y, w=w, deg=order, cov=True)
    pval = np.polyval(popt, 0.0)
    perr = np.sqrt(np.diag(pcov))[-1]
    return (pval - row[ycol]) ** 2 > metric * perr


def threshold(threshold, data, rising=True, falling=False):
    """Implement the threshold method - also used in peak-finder.

    Args:
        threshold (float):
            Threshold valuye in data to look for
        rising (bool):
            Find points where data is rising up past threshold
        falling (bool):
            Find points where data is falling below the threshold

    Returns:
        (array):
            Fractional indices where the data has crossed the threshold assuming a
            straight line interpolation between two points.
    """
    # First we find all points where we cross zero in the correct direction
    current = data
    previous = np.roll(current, 1)
    current = np.atleast_1d(current)
    index = np.arange(len(current))
    sdat = np.column_stack((index, current, previous))
    if rising and not falling:
        expr = lambda x: (x[1] >= threshold) & (x[2] < threshold)
    elif rising and falling:
        expr = lambda x: ((x[1] >= threshold) & (x[2] < threshold)) | ((x[1] <= threshold) & (x[2] > threshold))
    elif falling and not rising:
        expr = lambda x: (x[1] <= threshold) & (x[2] > threshold)
    else:
        expr = lambda x: False

    # Now we refine the estimate of zero crossing with a cubic interpolation
    # and use Newton's root finding method to locate the zero in the interpolated data

    intr = interp1d(index, data.ravel() - threshold, kind="cubic")
    roots = []
    for ix, x in enumerate(sdat):
        if ix > 0 and expr(x):  # There's a root somewhere here !
            try:
                roots.append(newton(intr, ix))
            except (ValueError, RuntimeError):  # fell off the end here
                pass
    return np.array(roots)


def _twoD_fit(xy1, xy2, xmode="linear", ymode="linear", m0=None):
    r"""Calculae an optimal transformation of points :math:`(x_1,y_1)\rightarrow(x_2,y_2)`.

    Arguments:
        xy1 ( n by 2 array of float):
            Set of points to be mapped from.
        xy2 ( n by 2 array of floats):
            Set of points to be mapped to.

    Keyword Arguments:
        xmode ('affine', 'linear', 'scale' 'offset' or 'fixed'):
            How to manipulate the x-data
        ymode ('linear', 'scale' 'offset' or 'fixed'):
            How to manipulate the y-data
        m0 (3x2 array):
            Initial and fixed values of the transformation. Defaults to using an identity transformation.

    Returns:
        (tuple of opt_trans,trans_err,mapping func)

    The most general case is an affine transform which includes rotation, scale, translation and skew. This is
    represented as a 2 x 3 matrix  of coordinates. The *xmode* and *ymode* parameters control the possible operations
    to align the data in x and y directions, in addition to which the *xmode* parameter can take the value 'affine'
    which allows a full affine transformation. The returned values are the affine transformation matrix, the
    uncertainties in this and a function to map coordinates with the optimal affine transformation.

    Note:
        *m0* combines both giving an initial value and fixed values for the transformation. If *m0* is set, then it
        is used to provide initial balues of the free parameters. Which elelemnts of *m0* that are free parameters
        and which are fixed is determined by the *xmode* and *ymode* parameters. IF *xmode* and *ymode* are both
        fixed, however, no scaling is done at all.
    """
    if xy1.shape != xy2.shape or xy1.shape[1] != 2:
        raise RuntimeError(f"coordinate arrays must be equal length with two columns, not {xy1.shape} and {xy2.shape}")
    xvarp = {
        "affine": [[0, 0], [0, 1], [0, 2], [1, 0], [1, 1], [1, 2]],
        "linear": [[0, 0], [0, 2]],
        "scale": [[0, 0]],
        "offset": [[0, 2]],
        "fixed": [[]],
    }
    yvarp = {"linear": [[1, 1], [1, 2]], "scale": [[1, 1]], "offset": [[1, 2]], "fixed": [[]]}

    if xmode not in xvarp or ymode not in yvarp:
        raise RuntimeError(f"xmode and ymode must be one of 'linear','scale','offset','fixed' not {xmode} and {ymode}")

    if xmode == "affine":
        ymode = "fixed"

    xunknowns = len(xvarp[xmode])
    yunknowns = len(yvarp[ymode])
    if xunknowns + yunknowns == 0:  # shortcircuit for the trivial case
        return np.array([[1, 0], [1, 0]]), np.zeros((2, 2)), lambda x: x

    mapping = xvarp[xmode] + yvarp[ymode]
    mapping = [m for m in mapping if m != []]  # remove empty mappings
    data = np.column_stack((xy1, xy2)).T

    if isinstance(m0, list):
        m0 = np.array(m0)

    if m0 is None:
        p0s = {"affine": [1, 0, 0, 0, 1, 0], "linear": [1, 0], "scale": [1], "offset": [0], "fixed": []}
        p0 = p0s[xmode] + p0s[ymode]
        default = np.array([[1.00, 0.0, 0.0], [0.0, 1.0, 0.0]])
    elif isinstance(m0, np.ndarray) and m0.shape == (2, 3):
        p0 = [0] * len(mapping)
        for i, [u, v] in enumerate(mapping):
            p0[i] = m0[u, v]
            default = m0
    else:
        raise RuntimeError(f"m0 starting matrix should be a numpy array of size (2,3) not {m0}")

    result = np.zeros(len(xy1))

    def transform(xy, *p):  # Construct the fitting function
        """Fitting function to find the transfoprm."""
        xy1 = np.column_stack((xy[:2, :].T, np.ones(xy.shape[1]))).T
        xy2 = xy[2:, :]
        for pi, (u, v) in zip(p, mapping):
            default[u, v] = pi
        xyt = np.dot(default, xy1)
        ret = np.sqrt(np.sum((xy2 - xyt) ** 2, axis=0))
        return ret

    popt, pcov = curve_fit(transform, data, result, p0=p0)
    perr = np.sqrt(np.diag(pcov))

    # Initialise the return values
    default = np.array([[1.00, 0.0, 0.0], [0.0, 1.0, 0.0]])
    for pi, (u, v) in zip(popt, mapping):
        default[u, v] = pi
    default_err = np.zeros((2, 3))
    for pi, (u, v) in zip(perr, mapping):
        default_err[u, v] = pi

    transform = lambda xy: ApplyAffineTransform(xy, default)

    return (default, default_err, transform)


def ApplyAffineTransform(xy, transform):
    """Apply a given afffine transform to a set of xy data points.

    Args:
        xy (n by 2 array):
            Set of x,y coordinates to be transformed
        transform (2 by 3 array):
            Affine transform matrix.

    Returns:
        (n x 2 array).
        Transformed coordinates.
    """
    xyt = np.row_stack((xy.T, np.ones(len(xy))))
    xyt = np.dot(transform, xyt)
    return xyt.T


def GetAffineTransform(p, pd):
    """Calculate an affine transform from 2 sets of three points.

    Args:
        p (3x2 array):
            Coordinates of points to transform from.
        pd (3x2 array):
            Coordinates of points to transform to.

    Returns:
        (2x3 array):
            matrix representing the affine transform.
    """
    if np.shape(p) != (3, 2) and np.shape(pd) != (3, 2):
        raise RuntimeError("Must supply three points")

    p = np.append(p, np.atleast_2d(np.ones(3)).T, axis=1)
    transform = np.linalg.solve(p, pd)
    return transform.T



"""Detect peaks in a dataset."""

# pylint: disable=invalid-name
from Stoner import Data
from Stoner.analysis.fitting.models.generic import Linear

from matplotlib.cm import jet
from numpy import linspace, log10

d = Data("../../sample-data/New-XRay-Data.dql")
d.y = log10(d.y)
d.smooth(
    windows="hanning", size=4, replace=True, result=True, padtype="constant"
)
d.setas = "x.y"
d.lmfit(Linear, residuals=True, result=True)
d.setas = "x...y"
d.data = d.data[4:-4]
d.plot()
d.find_peaks(
    modify=True, prominence=(0.05, 20), width=(0.02, 0.1), distance=0.01
)
d.labels[1] = "Find_Peaks"
d.plot(figure=d.fig, fmt="bx")
d.title = "Finding peaks with peaks() and find_peaks()"

# Use the metadata that find_peaks produces
colors = jet(linspace(0, 1, len(d)), alpha=0.5)
for l, r, c in zip(d["left_ips"], d["right_ips"], colors):
    ax = d.axes[d.ax]
    ax.axvspan(l, r, facecolor=c)



# -*- coding: utf-8 -*-
"""Loader for maximus spectra files."""
from pathlib import Path

import numpy as np

from ..decorators import register_loader
from ...core.exceptions import StonerLoadError
from ..utils.maximus import read_scan, flatten_header, hdr_to_dict


@register_loader(
    patterns=[(".hdr", 16), (".xsp", 16)], mime_types=("text/plain", 16), name="MaximusSpectra", what="Data"
)
def load_maximus_spectra(new_data, *args, **kargs):
    """Maximus xsp file loader routine.

    Args:
        filename (string or bool): File to load. If None then the existing filename is used,
            if False, then a file dialog will be used.

    Returns:
        A copy of the itnew_data after loading the data.
    """
    filename = kargs.get("filename", args[0])
    if filename is None or not filename:
        new_data.get_filename("r")
    else:
        new_data.filename = filename
    # Open the file and read the main file header and unpack into a dict
    try:
        pth = Path(new_data.filename)
    except (TypeError, ValueError) as err:
        raise StonerLoadError("Can only open things that can be converted to paths!") from err
    if pth.suffix != ".hdr":  # Passed a .xim or .xsp file in instead of the hdr file.
        pth = Path("_".join(str(pth).split("_")[:-1]) + ".hdr")
    stem = pth.parent / pth.stem

    try:
        hdr = flatten_header(hdr_to_dict(pth))
        if "Point Scan" not in hdr["ScanDefinition.Type"]:
            raise StonerLoadError("Not an Maximus Single Image File")
    except (StonerLoadError, ValueError, TypeError, IOError) as err:
        raise StonerLoadError("Error loading as Maximus File") from err
    header, data, dims = read_scan(stem)
    new_data.metadata.update(flatten_header(header))
    new_data.data = np.column_stack((dims[0], data))
    headers = [new_data.metadata["ScanDefinition.Regions.PAxis.Name"]]
    if len(dims) == 2:
        headers.extend([str(x) for x in dims[1]])
    else:
        headers.append(new_data.metadata["ScanDefinition.Channels.Name"])
    new_data.column_headers = headers
    new_data.setas = "xy"
    return new_data


@register_loader(
    patterns=[(".hdr", 16), (".xim", 16)], mime_types=("text/plain", 16), name="MaximusImage", what="Data"
)
def load_maximus_data(new_data, filename, *args, **kargs):
    """Load a maximus image, but to a Data object."""
    try:
        new_data.filename = filename
        pth = Path(new_data.filename)
    except TypeError as err:
        raise StonerLoadError(f"UUnable to interpret {filename} as a path like object") from err
    if pth.suffix != ".hdr":  # Passed a .xim or .xsp file in instead of the hdr file.
        pth = Path("_".join(str(pth).split("_")[:-1]) + ".hdr")
    stem = pth.parent / pth.stem

    try:
        hdr = flatten_header(hdr_to_dict(pth))
        if "Image Scan" not in hdr["ScanDefinition.Type"]:
            raise StonerLoadError("Not an Maximus Single Image File")
    except (StonerLoadError, ValueError, TypeError, IOError) as err:
        raise StonerLoadError("Error loading as Maximus File") from err
    data = read_scan(stem)[1]
    new_data.metadata.update(hdr)
    if data.ndim == 3:
        data = data[:, :, 0]
    new_data.data = data
    return new_data



"""Detect outlying points from a lione."""

# pylint: disable=invalid-name
import numpy as np

from Stoner import Data
from Stoner.analysis.utils import poly_outlier

np.random.seed(12345)
x = np.linspace(0, 100, 201)
y = 0.01 * x**2 + 5 * np.sin(x / 10.0)

i = np.random.randint(len(x) - 20, size=20) + 10
y[i] += np.random.normal(size=len(i), scale=20)

d = Data(np.column_stack((x, y)), column_headers=["x", "y"], setas="xy")
d.plot(fmt="b.", label="raw data")
e = d.clone
e.outlier_detection(window=5, action="delete")
e.plot(fmt="r-", label="Default Outliers removed")
h = d.clone
h.outlier_detection(window=5, action="delete", shape="hamming")
h.plot(color="orange", label="Default Outliers removed with Hamming window")
f = d.clone
f.outlier_detection(
    window=21, order=3, certainty=2, width=3, action="delete", func=poly_outlier
)
f.plot(fmt="g-", label="Poly Outliers removed")
g = d.clone
g = g.outlier_detection(
    window=21, order=3, certainty=3, width=3, action="delete", func=poly_outlier
)
g.plot(color="purple", label="Masked outliers")
g = d.clone
e.title = "Outlier detection test"



"""Example of nDimArrhenius Fit."""

# pylint: disable=invalid-name
from numpy import linspace, ones_like
from numpy.random import normal

from Stoner import Data
from Stoner.analysis.fitting.models.tunnelling import BDR, bdr

# Make some data
V = linspace(-10, 10, 1000)
I = bdr(V, 2.5, 3.2, 0.3, 15.0, 1.0) + normal(size=len(V), scale=1.0e-3)
dI = ones_like(V) * 1.0e-3

# Curve fit
d = Data(V, I, dI, setas="xye", column_headers=["Bias", "Current", "Noise"])

d.curve_fit(bdr, p0=[2.5, 3.2, 0.3, 15.0, 1.0], result=True, header="curve_fit")
d.setas = "xyey"
d.plot(fmt=["r.", "b-"])
d.annotate_fit(
    bdr,
    x=0.6,
    y=0.05,
    prefix="bdr",
    fontdict={"size": "x-small", "color": "blue"},
)

# lmfit
d.setas = "xy"
fit = BDR(missing="drop")
p0 = fit.guess(I, x=V)
for p, v, mi, mx in zip(
    ["A", "phi", "dphi", "d", "mass"],
    [2.500, 3.2, 0.3, 15.0, 1.0],
    [0.100, 1.0, 0.05, 5.0, 0.5],
    [10, 10.0, 2.0, 30.0, 5.0],
):
    p0[p].value = v
    p0[p].bounds = [mi, mx]
d.lmfit(fit, p0=p0, result=True, header="lmfit")
d.setas = "x...y"
d.plot(fmt="g-")
d.annotate_fit(
    fit,
    x=0.2,
    y=0.05,
    prefix="BDR",
    fontdict={"size": "x-small", "color": "green"},
)

d.ylabel = "Current"
d.title = "BDR Model test"



"""Demo of new Stoner.Analysis.AnalyseFile.lmfit."""

# pylint: disable=invalid-name
from os.path import join

from Stoner.analysis.fitting.models import cfg_data_from_ini, cfg_model_from_ini
from Stoner import __home__

config = join(__home__, "..", "scripts", "PCAR-New.ini")
datafile = join(__home__, "..", "sample-data", "PCAR Co Data.csv")

d = cfg_data_from_ini(config, datafile)
model, p0 = cfg_model_from_ini(config, data=d)

fit = d.lmfit(model, result=True, header="lmfit", output="report")

d.plot_xy(0, [2, 1], ["r-", "bo"])  # plot the data
d.annotate_fit(model, x=0.05, y=0.75, fontdict={"size": "x-small"})

print(fit.fit_report())



"""Sub-package of routines to load ImageFile objects in different formats.

This sub-package is lazy-loaded but then pulls in all of the modules in order to register the routines.
"""

__all__ = ["generic", "hdf5", "facilities", "maximus"]

from . import generic, hdf5, facilities, maximus



# -*- coding: utf-8 -*-
"""
Created on Fri May 27 17:09:04 2016

@author: phyrct
"""

from Stoner.Image import ImageArray, ImageFile
from Stoner.Core import typeHintedDict
from Stoner import Data, __home__
import numpy as np
import pytest
from os import path
import tempfile
import os
import sys
import shutil
from PIL import Image

from scipy.version import version as spv

spv = [int(x) for x in spv.split(".")]


# data arrays for testing - some useful small images for tests

thisdir = path.dirname(__file__)


def shares_memory(arr1, arr2):
    """Check if two numpy arrays share memory"""
    # ret = arr1.base is arr2 or arr2.base is arr1
    ret = np.may_share_memory(arr1, arr2)
    return ret


def _has_method(obj, method):
    """Returns True if obj has a method."""
    return callable(getattr(obj, method, None))


# random array shape 3,4
selfarr = np.array(
    [
        [0.41674764, 0.66100043, 0.91755303, 0.33796703],
        [0.06017535, 0.1440342, 0.34441777, 0.9915282],
        [0.2984083, 0.9167951, 0.73820304, 0.7655299],
    ]
)
selfimarr = ImageArray(np.copy(selfarr))  # ImageArray object
selfimarrfile = ImageArray(os.path.join(thisdir, "coretestdata/im1_annotated.png"))
# ImageArray object from file

selfimarr_x = np.copy(selfimarr).view(ImageArray)
#####test loading with different datatypes  ####


def test_load_from_array():
    # from array
    assert np.array_equal(selfimarr, selfarr)
    # int type
    imarr = ImageArray(np.arange(12, dtype="int32").reshape(3, 4))
    assert imarr.dtype == np.dtype("int32"), "Failed to set correct dtype - actual dtype={}".format(imarr.dtype)


def test_load_from_ImageArray():
    # from ImageArray
    t = ImageArray(selfimarr)
    assert shares_memory(selfimarr, t), "no overlap on creating ImageArray from ImageArray"


def test_load_from_png():
    subpath = os.path.join("coretestdata", "im1_annotated.png")
    fpath = os.path.join(thisdir, subpath)
    anim = ImageArray(fpath)
    assert (
        os.path.normpath(anim.metadata["Loaded from"]).lower() == os.path.normpath(fpath).lower()
    ), "Failed with {os.path.normpath(anim.metadata['Loaded from'])} and {os.path.normpath(fpath)}"
    cwd = os.getcwd()
    os.chdir(thisdir)
    anim = ImageArray(subpath)
    # check full path is in loaded from metadata
    assert (
        os.path.normpath(anim.metadata["Loaded from"]).lower() == os.path.normpath(fpath).lower()
    ), "Full path not in metadata: {}".format(anim["Loaded from"])
    os.chdir(cwd)


def test_load_save_all():
    tmpdir = tempfile.mkdtemp()
    pth = path.join(__home__, "..")
    datadir = path.join(pth, "sample-data")
    image = ImageFile(path.join(datadir, "kermit.png"))
    ims = {}
    fmts = ["uint8", "uint16", "uint32", "float32"]
    modes = {"uint8": "L", "uint32": "RGBX", "float32": "F"}
    for fmt in fmts:
        ims[fmt] = image.clone.convert(fmt)
        ims[fmt].save_tiff(path.join(tmpdir, "kermit-{}.tiff".format(fmt)))
        ims[fmt].save_tiff(path.join(tmpdir, "kermit-forcetype{}.tiff".format(fmt)), forcetype=True)
        ims[fmt].save_npy(path.join(tmpdir, "kermit-{}.npy".format(fmt)))
        if fmt != "uint16":
            im = Image.fromarray((ims[fmt].image.view(np.ndarray)).astype(fmt), mode=modes[fmt])
            im.save(path.join(tmpdir, "kermit-nometadata-{}.tiff".format(fmt)))
        del ims[fmt]["Loaded from"]
    for fmt in fmts:
        iml = ImageFile(path.join(tmpdir, "kermit-{}.tiff".format(fmt)))
        del iml["Loaded from"]
        assert ims[fmt] == iml, f"{ims[fmt].metadata} {iml.metadata}"
        iml = ImageFile(path.join(tmpdir, "kermit-{}.npy".format(fmt)))
        del iml["Loaded from"]
        assert np.all(ims[fmt].data == iml.data), "Round tripping npy with format {} failed".format(fmt)
        if fmt != "uint16":
            im = ImageFile(path.join(tmpdir, "kermit-nometadata-{}.tiff".format(fmt)))
            assert np.all(im.data == ims[fmt].data), "Loading from tif without metadata failed for {}".format(fmt)
    shutil.rmtree(tmpdir)
    _ = image.convert("uint8")


def test_conversion_bitness():
    pth = path.join(__home__, "..")
    datadir = path.join(pth, "sample-data")
    image = ImageFile(path.join(datadir, "kermit.png"))
    img = image.clone
    img2 = img.clone // 256
    img.convert("uint8")  # Down conversion with prevision loss
    assert np.all(img.image == img2.image)
    img.convert("int16")  # Upconversion with unequal bit lengths
    assert img.max() == 32767 and img.min() == 0
    img3 = img2.clone.convert("uint8")  # Downconversion without precision loss
    assert np.all(img2.image == img3.image)


def test_load_from_ImageFile():
    # uses the ImageFile.im attribute to set up ImageArray. Memory overlaps
    imfi = ImageFile(selfarr)
    imarr = ImageArray(imfi)
    assert np.array_equal(imarr, imfi.image), "Initialising from ImageFile failed"
    assert shares_memory(imarr, imfi.image)


def test_load_from_list():
    t = ImageArray([[1, 3], [3, 2], [4, 3]])
    assert np.array_equal(t, np.array([[1, 3], [3, 2], [4, 3]])), "Initialising from list failed"


def test_load_1d_data():
    t = ImageArray(np.arange(10) / 10.0)
    assert len(t.shape) == 2  # converts to 2d


def test_load_no_args():
    # Should be a 2d empty array
    t = ImageArray()
    assert len(t.shape) == 2
    assert t.size == 0


def test_load_bad_data():
    def testload(arg):
        ImageArray(arg)

    # dictionary
    with pytest.raises(ValueError):
        testload({"a": 1})

    # bad filename
    with pytest.raises(ValueError):
        testload("sillyfile.xyz")


def test_load_kwargs():
    # metadata keyword arg
    t = ImageArray(selfarr, metadata={"a": 5, "b": 7})
    assert ("a" in t.metadata.keys()) and ("b" in t.metadata.keys())
    assert t.metadata["a"] == 5
    # asfloat
    t = ImageArray(np.arange(12).reshape(3, 4), asfloat=True)
    assert t.dtype == np.float64, "Initialising asfloat failed"
    assert "Loaded from" in t.metadata.keys(), "Loaded from should always be in metadata"


# ####test attributes ##


def test_filename():
    im = ImageArray(np.linspace(0, 1, 12).reshape(3, 4))
    fpath = os.path.join(thisdir, "coretestdata/im1_annotated.png")
    assert (
        os.path.normpath(selfimarrfile.filename).lower() == os.path.normpath(fpath).lower()
    ), f"Failed with {os.path.normpath(selfimarrfile.filename)} and {os.path.normpath(fpath)}"
    im = ImageArray(np.linspace(0, 1, 12).reshape(3, 4))
    im["Loaded from"]
    im.filename
    assert im.filename == "", "{}, {}".format(selfimarr.shape, im.filename)


def test_clone():
    selfimarr["abc"] = 123  # add some metadata
    selfimarr["nested"] = [1, 2, 3]  # add some nested metadata to check deepcopy
    selfimarr.userxyz = 123  # add a user attribute
    c = selfimarr.clone
    assert isinstance(c, ImageArray), "Clone not ImageArray"
    assert np.array_equal(c, selfimarr), "Clone not replicating elements"
    assert all([k in c.metadata.keys() for k in selfimarr.metadata.keys()]), "Clone not replicating metadata"
    assert not shares_memory(c, selfimarr), "memory overlap on clone"  # formal check
    selfimarr["bcd"] = 234
    assert "bcd" not in c.metadata.keys(), "memory overlap for metadata on clone"
    selfimarr["nested"][0] = 2
    assert selfimarr["nested"][0] != c["nested"][0], "deepcopy not working on metadata"
    assert c.userxyz == 123
    c.userxyz = 234
    assert c.userxyz != selfimarr.userxyz
    img = ImageFile(np.zeros((100, 100)))
    img.abcd = "Hello"
    img2 = img.clone
    assert img2.abcd == "Hello"
    del img["Loaded from"]
    assert len(img.metadata) == 0
    del img.abcd
    assert not hasattr(img, "abcd")


def test_metadata():
    assert isinstance(selfimarr.metadata, typeHintedDict)
    selfimarr["testmeta"] = "abc"
    assert selfimarr["testmeta"] == "abc", "Couldn't change metadata"
    del selfimarr["testmeta"]
    assert "testmeta" not in selfimarr.keys(), "Couldn't delete metadata"

    # bad data
    def test(imarr):
        imarr.metadata = (1, 2, 3)

    with pytest.raises(TypeError):
        test(selfimarr)  # check it won't let you do this


### test numpy like creation behaviour #
def test_user_attributes():
    selfimarr.abc = "new att"
    assert hasattr(selfimarr, "abc")
    t = ImageArray(selfimarr)
    assert hasattr(t, "abc"), "problem copying new attributes"
    t = selfimarr.view(ImageArray)  # check array_finalize copies attribute over
    assert hasattr(t, "abc")
    t = selfimarr * np.random.random(selfimarr.shape)
    assert isinstance(t, ImageArray), "problem with ufuncs"
    assert hasattr(t, "abc"), "Ufunc lost attribute!"


#####  test functionality  ##
def test_save():
    testfile = path.join(thisdir, "coretestdata", "testsave")
    ext = [".png", ".npy"]
    keys = selfimarr.keys()
    for e in ext:
        selfimarr.save(testfile + e)
        load = ImageArray(testfile + e)
        assert all([k in load.keys() for k in keys]), "problem saving metadata {} {}".format(list(load.keys()), e)
        if e == ".npy":
            # tolerance is really poor for png which saves in 8bit format
            assert np.allclose(selfimarr, load), "data not the same for extension {}".format(e)
        os.remove(testfile + e)  # tidy up


def test_savetiff():
    testfile = path.join(thisdir, "coretestdata", "testsave.tiff")
    # create a few different data types
    testb = ImageArray(np.zeros((4, 5), dtype=bool))  # bool
    testb[0, :] = True
    testui = ImageArray(np.arange(20).reshape(4, 5))  # int32
    testi = ImageArray(np.copy(testui) - 10)
    testf = ImageArray(np.linspace(-1, 1, 20).reshape(4, 5))  # float64
    for im in [testb, testui, testi, testf]:
        im["a"] = [1, 2, 3]
        im["b"] = "abc"  # add some test metadata
        im.filename = testfile
        im.save()
        n = ImageArray(testfile)
        assert all([n["a"][i] == im["a"][i] for i in range(len(n["a"]))])
        assert n["b"] == im["b"]
        assert "ImageArray.dtype" in n.metadata.keys()  # check the dtype metadata got added
        assert im.dtype == n.dtype  # check the datatype
        assert np.allclose(im, n)  # check the data


def test_max_box():
    s = selfimarr.shape
    assert selfimarr.max_box == (0, s[1], 0, s[0])


def test_crop():
    c = selfimarr.crop((1, 3, 1, 4), copy=True)
    assert np.array_equal(c, selfimarr[1:4, 1:3]), "crop didn't work"
    assert not shares_memory(c, selfimarr), "crop copy failed"
    c2 = selfimarr.crop(1, 3, 1, 4, copy=True)
    assert np.array_equal(c2, c), "crop with separate arguments didn't work"
    c3 = selfimarr.crop(box=(1, 3, 1, 4), copy=False)
    assert np.array_equal(c3, c), "crop with no arguments failed"
    assert shares_memory(selfimarr, c3), "crop with no copy failed"


def test_asint():
    ui = selfimarr.asint()
    assert ui.dtype == np.uint16
    intarr = np.array(
        [[27311, 43318, 60131, 22148], [3943, 9439, 22571, 64979], [19556, 60082, 48378, 50169]], dtype=np.uint16
    )
    assert np.array_equal(ui, intarr)


def test_other_funcs():
    """test imagefuncs add owns. the functions themselves are not checked
    and should include a few examples in the doc strings for testing"""
    assert hasattr(selfimarr, "do_nothing"), "imagefuncs not being added to dir"
    assert hasattr(
        selfimarr, "Stoner__Image__imagefuncs__do_nothing"
    ), "Stoner image functions not added with full namke"
    assert hasattr(selfimarr, "img_as_float"), "skimage funcs not being added to dir"
    im = selfimarr.do_nothing()  # see if it can run
    assert np.allclose(im, selfimarr), "imagefuncs not working"
    assert shares_memory(im, selfimarr), "imagefunc failed to share memory"
    im0 = ImageArray(np.linspace(0, 1, 12).reshape(3, 4))
    im1 = im0.clone * 5
    im2 = im1.rescale_intensity()  # test skimage
    assert np.allclose(im2, im0), "skimage func failed"
    assert not shares_memory(im2, im1), "skimage failed to clone"
    im3 = im1.skimage__exposure__exposure__rescale_intensity()  # test call with module name
    assert np.allclose(im3, im0), "skimage call with module name failed"


def test_attrs():
    test = ImageArray([])
    assert _has_method(test, "gaussian_filter"), "Failed to get scipy.ndimage.gaussian_filter as ImageArray attr"
    assert _has_method(test, "gaussian"), "Failed to get skimage.filter.gaussian as ImageArray attr"
    assert _has_method(test, "gridimage"), "Failed to get imagfuncs.gridimage has ImageArray attr"


selfa = np.linspace(0, 5, 12).reshape(3, 4)
selfifi = ImageFile(selfa)
selfimgFile = ImageFile(os.path.join(thisdir, "coretestdata/im1_annotated.png"))


def test_constructors():
    selfimgFile = ImageFile(os.path.join(thisdir, "coretestdata/im1_annotated.png"))
    selfd = Data(selfimgFile)
    selfimgFile2 = ImageFile(selfd)
    selfimgFile2.pop("Stoner.class", "")
    selfimgFile.pop("Stoner.class", "")
    del selfimgFile2["x_vector"]
    del selfimgFile2["y_vector"]
    assert np.all(
        selfimgFile.image == selfimgFile2.image
    ), "Roundtripping constructor through Data failed to dupliocate data."
    assert (
        selfimgFile.metadata == selfimgFile2.metadata
    ), "Roundtripping constructor through Data failed to duplicate metadata: {} {}".format(
        selfimgFile2.metadata ^ selfimgFile.metadata, selfimgFile.metadata ^ selfimgFile2.metadata
    )


def test_properties():
    selfa = np.linspace(0, 5, 12).reshape(3, 4)
    assert np.allclose(selfifi.image, selfa)
    selfifi[0, 1] = 10.1
    assert selfifi.image[0, 1] == 10.1


def test_attrs2():
    """Check that creating new attributes puts them on ImageFile and not ImageArray and updates public_attrs."""
    assert selfifi["Loaded from"] == ""
    selfifi.abc = 123
    assert selfifi.abc == 123
    assert "abc" in selfifi.__dict__
    assert "abc" not in selfifi._image.__dict__
    assert "abc" in selfifi._public_attrs


def test_methods():
    b = np.arange(12).reshape(3, 4)
    ifi = ImageFile(b)
    ifi.asfloat(normalise=False, clip_negative=False)  # convert in place
    assert ifi.image.dtype.kind == "f"
    assert np.max(ifi) == 11.0
    ifi.image == ifi.image * 5
    ifi.rescale_intensity()
    assert np.allclose(ifi.image, np.linspace(0, 1, 12).reshape(3, 4))
    ifi.crop(0, 3, 0, None)
    assert ifi.shape == (3, 3)  # check crop is forced to overwrite ifi despite shape change
    datadir = path.join(__home__, "..", "sample-data")
    image = ImageFile(path.join(datadir, "kermit.png")).asfloat(normalise=False)
    i2 = image.clone.crop(5, _=True)
    assert i2.shape == (469, 349), "Failed to trim box by integer"
    i2 = image.clone.crop(0.25, _=True)
    assert i2.shape == (269, 269), "Failed to trim box by float"
    i2 = image.clone
    i2.crop([0.1, 0.2, 0.05, 0.1], _=True)
    assert i2.shape == (24, 36), "Failed to trim box by sequence of floats"
    assert image.aspect == pytest.approx(0.7494780793), "Aspect ratio failed"
    assert image.centre == (239.5, 179.5), "Failed to read image.centre"
    i2 = image.CW
    assert i2.shape == (359, 479), "Failed to rotate clockwise"
    i3 = i2.CCW
    assert i3.shape == (479, 359), "Failed to rotate counter-clockwise"
    i3 = i2.flip_h
    assert np.all(i3[:, 0] == i2[:, -1]), "Flip Horizontal failed"
    i3 = i2.flip_v
    assert np.all(i3[0, :] == i2[-1, :]), "Flip Horizontal failed"
    i2 = image.clone
    i3 = i2 - 127
    assert i3.mean() == pytest.approx(33940.72596111909, rel=1e-2), "Subtract integer failed."
    with pytest.raises(TypeError):
        i2 - "Gobble"
    test = i2
    assert _has_method(test, "gaussian_filter"), "Failed to get scipy.ndimage.gaussian_filter as ImageFile attr"
    assert _has_method(test, "gaussian"), "Failed to get skimage.filter.gaussian as ImageFile attr"
    assert _has_method(test, "gridimage"), "Failed to get imagfuncs.gridimage has ImageFile attr"
    assert image._repr_png_().startswith(b"\x89PNG\r\n"), "Failed to do ImageFile png representation"


def test_mask():
    i = np.ones((200, 200), dtype="uint8") * np.linspace(1, 200, 200).astype("uint8")
    i = ImageFile(i)
    i.mask.draw.rectangle(100, 50, 100, 100)
    assert i.mean() == pytest.approx(117.1666666666666, 1.0), "Mean after masked rectangle failed"
    i.mask.invert()
    assert i.mean() == pytest.approx(50.5, 1.0), "Mean after inverted masked rectangle failed"
    i.mask.clear()
    assert i.mean() == pytest.approx(100.5, 1.0), "Mean after clearing mask failed"
    i.mask.draw.square(100, 50, 100)
    assert i.mean() == pytest.approx(117.1666666666666, 1.0), "Mean after masked rectangle faile"
    i.mask.clear()
    i.mask.draw.annulus(100, 50, 35, 25)
    assert i.mean() == pytest.approx(102.96850393700, 1.0), "Mean after annular block mask failed"
    i.mask = False
    i.mask.draw.annulus(100, 50, 25, 35)
    assert i.mean() == pytest.approx(51.0), "Mean after annular pass mask failed"
    i.mask[:, :] = False
    assert not np.any(i.mask), "Setting Mask by index failed"
    i.mask = -i.mask
    assert np.all(i.mask), "Setting Mask by index failed"
    i.mask = ~i.mask
    assert not np.any(i.mask), "Setting Mask by index failed"
    i.mask.draw.circle(100, 100, 20)
    st = repr(i.mask)
    assert st.count("X") == i.mask.sum(), "String representation of mask failed"
    assert np.all(i.mask.image == i.mask._mask), "Failed to access mak data by image attr"
    i.mask = False
    i2 = i.clone
    i.mask.draw.rectangle(100, 100, 100, 50, angle=np.pi / 2)
    i2.mask.draw.rectangle(100, 100, 50, 100)
    assert np.all(i.mask.image == i2.mask.image), "Drawing rectangle with angle failed"
    assert i.mask._repr_png_().startswith(b"\x89PNG\r\n"), "Failed to do mask png representation"
    i = ImageFile(np.zeros((100, 100)))
    i2 = i.clone
    i2.draw.circle(50, 50, 25)
    i.mask = i2
    assert i.mask.sum() == i2.sum(), "Setting mask from ImageFile Failed"
    i2.mask = i.mask
    assert np.all(i.mask.image == i2.mask.image), "Failed to set mask by mask proxy"
    i = ImageFile(np.ones((100, 100)))
    i.mask.draw.square(50, 50, 10)
    i.mask.rotate(angle=np.pi / 4)
    i.mask.invert()
    i2 = ImageFile(np.zeros((100, 100)))
    i2.draw.square(50, 50, 10, angle=np.pi / 4)
    assert i.sum() == pytest.approx(i2.sum(), 1.5), "Check on rotated mask failed !"


def test_draw():
    i = ImageFile(np.zeros((200, 200)))
    attrs = [x for x in dir(i.draw) if not x.startswith("_")]
    expected = 21
    assert len(attrs) == expected, "Directory of DrawProxy failed"
    i2 = i.clone
    i2.draw.circle(10, 100, 10, value=1)
    assert i2.image.sum() == 305
    i2 = i.clone.draw.circle_perimeter_aa(100, 100, 10, value=1)
    assert np.isclose(i2.image.sum(), 56.5657137141714)
    i2 = i.clone.draw.random_shapes(i.shape, 1, 1, 30, 100)
    assert i2.label()[1] == 1
    i2 = i.clone.draw.rectangle_perimeter(100, 100, 20, 10, value=1)
    assert i2.image.sum() == 60.0
    i2.mask.threshold(0.5)
    assert i2.mask.sum() == 60


def test_operators():
    i = ImageFile(np.zeros((10, 10)))
    i = i + 1
    i += 4
    i += i.clone
    i += i.clone.data
    assert i.sum() == 2000, "Addition operators failed"
    i /= 4
    assert i.sum() == 500, "Division operators failed"
    i = i / 5
    assert i.sum() == 100, "Division operators failed"
    i2 = i.clone
    i -= 0.75
    i2 -= 0.25
    i3 = i2 // i
    assert i3.sum() == 50, "Division operators failed"
    i = ImageFile(np.zeros((10, 5)))
    i = ~i
    assert i.shape == (5, 10), "Invert to rotate failed"
    i.image = i.image.astype("uint8")
    i = -i
    assert i.sum() == 50 * 255, "Negate operators failed"


if __name__ == "__main__":  # Run some tests manually to allow debugging
    pytest.main([__file__, "--pdb"])



# -*- coding: utf-8 -*-
"""Demonstrate remove duplicates."""
# pylint: disable=no-member
from Stoner import Data, __homepath__

# Load data
data = Data(
    __homepath__ / ".." / "sample-data" / "TDI_Format_RT.txt", setas="xye"
)
data.e /= 10000
# Plot the original data
data.plot(fmt="k-", label="Original Data", capsize=3)
# De-dupe the data
data.remove_duplicates(strategy="average")
# Update the plot
data.plot(fmt="r.", label="De-duped data", capsize=3)
data.xlim(8.6, 10.0)  # pylint: disable=not-callable
data.ylim(7.4, 9.3)  # pylint: disable=not-callable
